/*
 * text.c
 *
 * Copyright 2022 Jonathan Schöbel <jonathan@Ubermos-2019>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */


#include <errno.h>
#include <math.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "macro.h"
#include "error.h"
#include "log.h"

#include "text.h"


struct SH_Text *
SH_Text_new (struct SH_Error * error)
{
	struct SH_Text * text;
	text = malloc (sizeof (struct SH_Text));

	if (text == NULL)
	{
		ERROR1 ("Memory allocation for SH_Text failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		return NULL;
	}

	text->length = 0;
	text->text = malloc (CHUNK_SIZE * sizeof (char));

	if (text->text == NULL)
	{
		ERROR1 ("Memory allocation for SH_Text failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		return NULL;
	}

	text->text[0] = 0;
	text->size = CHUNK_SIZE;

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return text;
}

void
SH_Text_free (struct SH_Text * text, struct SH_Error * error)
{
	free (text->text);

	free (text);

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return;
}

bool
SH_Text_enlarge (struct SH_Text * text, size_t size,
		 struct SH_Error * error)
{
	size_t new_size;

	if (size <= text->size)
	{
		if (error != NULL)
		{
			error->type = SUCCESS;
		}
		return TRUE;
	}

	/* if both operands were integers,
	 * this operation would fail
	 * as the division will behave like
	 * calling floor in the first place. */
	new_size = ceilf (size / (float) CHUNK_SIZE);
	if (new_size >= (SIZE_MAX / (float) CHUNK_SIZE))
	{
		ERROR1 ("Maximum length of SH_Text reached.\n");

		if (error != NULL)
		{
			error->type = DOMAIN_ERROR;
		}

		return FALSE;
	}

	new_size *= CHUNK_SIZE;
	text->text = realloc (text->text, new_size * sizeof (char));

	if (errno == ENOMEM)
	{
		ERROR1 ("Memory allocation for SH_Text data failed.\n");

		if (error != NULL)
		{
			error->type = ALLOCATION_FAILED;
		}

		return FALSE;
	}

	text->size = new_size;

	if (error != NULL)
	{
		error->type = SUCCESS;
	}

	return TRUE;
};

bool
SH_Text_append_string (struct SH_Text * text, char * string,
		       struct SH_Error * error)
{
	unsigned int length;

	length = strlen (string);

	if (!SH_Text_enlarge (text, text->length + length + 1, error))
	{
		return FALSE;
	}

	text->text = strcat (text->text, string);

	/* this is save as text->size is always at least one greater
	 * and SH_Text_enlarge hasn't failed. */
	text->length += length;

	return TRUE;
}

bool
SH_Text_append_text (struct SH_Text * text, struct SH_Text * text2,
		     struct SH_Error * error)
{
	if (!SH_Text_enlarge (text, text->length + text2->length + 1,
			      error))
	{
		return FALSE;
	}

	text->text = strcat (text->text, text2->text);

	/* this is save as text->size is always at least one greater
	 * and SH_Text_enlarge hasn't failed. */
	text->length += text2->length;

	return TRUE;
}

bool
SH_Text_join (struct SH_Text * text, struct SH_Text * text2,
	      struct SH_Error * error)
{
	if (!SH_Text_append_text (text, text2, error))
	{
		return FALSE;
	}

	SH_Text_free (text2, error);

	return TRUE;
}
